import unittest
import os

from tap_tester import LOGGER

potential_paths = [
    'tests/',
    '../tests/'
    'tap-stripe/tests/',
    '../tap-stripe/tests/',
]


def go_to_tests_directory():
    for path in potential_paths:
        if os.path.exists(path):
            os.chdir(path)
            return os.getcwd()
    raise NotImplementedError("This check cannot run from {}".format(os.getcwd()))

##########################################################################
### TEST
##########################################################################

class TestingTests(unittest.TestCase):
    def test_regression_suite(self):
        LOGGER.info("Acquiring path to tests directory.")
        cwd = go_to_tests_directory()

        LOGGER.info("Reading in filenames from tests directory.")
        files_in_dir = os.listdir(cwd)

        LOGGER.info("Dropping files that are not of the form 'test_<feature>.py'.")
        test_files_in_dir = [fn for fn in files_in_dir if fn.startswith('test_') and fn.endswith('.py')]

        LOGGER.info("Files found: %s", test_files_in_dir)

        LOGGER.info("Reading contents of circle config.")
        with open(cwd + "/../.circleci/config.yml", "r") as config:
            contents = config.read()

        LOGGER.info("Parsing circle config for run blocks.")
        runs = contents.replace(' ', '').replace('\n', '').split('-run_integration_test:')

        LOGGER.info("Verify all test files are executed in circle...")
        tests_not_found = set(test_files_in_dir)
        for filename in test_files_in_dir:
            LOGGER.info("\tVerifying %s is running in circle.", filename)
            trimmed_file_name = filename.split('.')[0].split('test_')[1]
            file_param = f"file:{trimmed_file_name}"
            if any([file_param in run for run in runs]):
                tests_not_found.remove(filename)
        self.assertSetEqual(tests_not_found, set())
        LOGGER.info("\t SUCCESS: All tests are running in circle.")




import os
from pathlib import Path
from random import random
from time import sleep, perf_counter
from datetime import datetime as dt
from dateutil.parser import parse

from collections import namedtuple

from tap_tester import menagerie, runner, connections, LOGGER
from base import BaseTapTest
from utils import \
    create_object, delete_object, list_all_object, stripe_obj_to_dict


# BUG_12478 | https://jira.talendforge.org/browse/TDL-12478
#             Fields that are consistently missing during replication
#             Original Ticket [https://stitchdata.atlassian.net/browse/SRCE-4736]
KNOWN_MISSING_FIELDS = {
    'customers': {
        'default_currency',
    },
    'subscriptions': {
        'automatic_tax',
        'cancellation_details',
        'default_tax_rates',
        'on_behalf_of',
        'payment_settings',
        'pending_update',
        'trial_settings',
    },
    'products': set(),
    'invoice_items': {
        'price',
    },
    'payouts': {
        'reconciliation_status',
    },
    'charges': set(),
    'subscription_items': set(),
    'plans': set(),
    'invoice_line_items': set(),
    'invoices': {
        'amount_shipping',
        'effective_at',
        'from_invoice',
        'latest_revision',
        'shipping_cost',
        'shipping_details',
    },
    'payment_intents': set()
}

# we have observed that the SDK object creation returns some new fields intermittently, which are not present in the schema
SCHEMA_MISSING_FIELDS = {
    'customers': {
        'test_clock'
    },
    'subscriptions': {
        'application',
        'currency',
        'description',
        'test_clock',
    },
    'products': {
        'default_price'
    },
    'invoice_items': {
        'test_clock',
    },
    'payouts': set(),
    'charges': {
        'failure_balance_transaction'
    },
    'subscription_items': set(),
    'plans': set(),
    'invoice_line_items': {
        'unit_amount_excluding_tax',
        'amount_excluding_tax'
    },
    'invoices': {
        'application',
        'rendering_options',
        'subtotal_excluding_tax',
        'test_clock',
        'total_excluding_tax'
    },
    'payment_intents': {
        'amount_details',
        'latest_charge'
    }
}

# `updated_by_event_type` field's value available in the records only if records are emitted by `event_updates`.
FIELDS_TO_NOT_CHECK = {
    'customers': {
        # Below fields are deprecated or renamed.(https://stripe.com/docs/upgrades#2019-10-17, https://stripe.com/docs/upgrades#2019-12-03)
        'account_balance',
        'tax_info',
        'tax_info_verification',
        'cards',
        'default_card',
        'updated_by_event_type'
    },
    'subscriptions': {
        # Below fields are deprecated or renamed.(https://stripe.com/docs/upgrades#2019-10-17, https://stripe.com/docs/upgrades#2019-12-03, https://stripe.com/docs/upgrades#2020-08-27)
        'billing',
        'start',
        'tax_percent',
        'invoice_customer_balance_settings',
        'updated_by_event_type'
    },
    'products': {
        # Below fields are available in the product record only if the value of the type field is `service`.
        # But, currently, during crud operation in all_fields test case, it creates product records of type `good`.
        'statement_descriptor',
        'unit_label',
        'updated_by_event_type'
    },
    'coupons': {
        # Field is not available in stripe documentation and also not returned by API response.(https://stripe.com/docs/api/coupons/object)
        'percent_off_precise',
        'updated_by_event_type'
    },
    'invoice_items': {
        'updated_by_event_type'
    },
    'payouts': {

        # Following fields are not mentioned in the documentation and also not returned by API (https://stripe.com/docs/api/payouts/object)
        'statement_description',
        'transfer_group',
        'source_transaction',
        'bank_account',
        'date',
        'amount_reversed',
        'recipient',
        'updated_by_event_type'
    },
    'charges': {
        # Following both fields `card` and `statement_description` are deprecated. (https://stripe.com/docs/upgrades#2015-02-18, https://stripe.com/docs/upgrades#2014-12-17)
        'card',
        'statement_description',
        'updated_by_event_type'
    },
    'subscription_items': {
        # Field is not available in stripe documentation and also not returned by API response. (https://stripe.com/docs/api/subscription_items/object)
      'current_period_end',
      'customer',
      'trial_start',
      'discount',
      'start',
      'tax_percent',
      'livemode',
      'application_fee_percent',
      'status',
      'trial_end',
      'ended_at',
      'current_period_start',
      'canceled_at',
      'cancel_at_period_end'
    },
    'invoices': {
        # Below fields are deprecated or renamed.(https://stripe.com/docs/upgrades#2019-03-14, https://stripe.com/docs/upgrades#2019-10-17, https://stripe.com/docs/upgrades#2018-08-11
        # https://stripe.com/docs/upgrades#2020-08-27)
        'application_fee',
        'billing',
        'closed',
        'date',
        # This field is deprcated in the version 2020-08-27
        'finalized_at',
        'forgiven',
        'tax_percent',
        'statement_description',
        'payment'
        'paid_out_of_band',
        'updated_by_event_type'
    },
    'plans': {
        # Below fields are deprecated or renamed. (https://stripe.com/docs/upgrades#2018-02-05, https://stripe.com/docs/api/plans/object)
        'statement_descriptor',
        'statement_description',
        'name',
        'updated_by_event_type',
        'tiers' # Field is not returned by API
    },
    'invoice_line_items': {
        # As per stripe documentation(https://stripe.com/docs/api/invoices/line_item#invoice_line_item_object-subscription_item),
        # 'subscription_item' is field that generated invoice item. It does not replicate in response if the line item is not an explicit result of a subscription.
        # So, due to uncertainty of this field, skipped it.
        'subscription_item',
        # As per stripe documentation(https://stripe.com/docs/api/invoices/line_item#invoice_line_item_object-invoice_item),
        # 'invoice_item' is id of invoice item associated wih this line if any. # So, due to uncertainty of this field, skipped it.
        'invoice_item'
    },
    'payment_intents': {
        'charges'
    }
}

KNOWN_FAILING_FIELDS = {
    'coupons': {
        'percent_off', # BUG_9720 | Decimal('67') != Decimal('66.6') (value is changing in duplicate records)
    },
    'customers': {
        # missing subfield 'rendering_options
        'invoice_settings'
    },
    'subscriptions': {
        # BUG_12478 | missing subfields in coupon where coupon is subfield within discount
        # BUG_12478 | missing subfields on discount ['checkout_session', 'id', 'invoice', 'invoice_item', 'promotion_code']
        'discount',
        # BUG_12478 | missing subfields on plan ['statement_description', 'statement_descriptor', 'name', 'amount_decimal']
        'plan',
    },
    'products': set(),
    'invoice_items': {
        'plan', # BUG_12478 | missing subfields
    },
    'payouts': set(),
    'charges': {
        # missing subfield ['card.mandate']
        'payment_method_details'
    },
    'subscription_items': {
        # BUG_12478 | missing subfields on plan ['statement_description', 'statement_descriptor', 'name']
        'plan',
        # missing subfields on price ['recurring.trial_period_days']
        'price'
    },
    'invoices': {
        'plans', # BUG_12478 | missing subfields
    },
    'plans': set(),
    'payment_intents':{
        # missing subfield ['payment_method_details.card.mandate']
        'charges',
        # missing subfield ['card.mandate_options']
        'payment_method_options',
        # missing subfield ['payment_method']
        'last_payment_error'
    },
    'invoice_line_items': {
        # missing subfield ['custom_unit_amount]
        'price'
    }
    # 'invoice_line_items': { # TODO This is a test issue that prevents us from consistently passing
    #     'unique_line_item_id',
    #     'invoice_item',
    # }
}

# NB | The following sets not to be confused with the sets above documenting BUGs.
#      These are testing issues/limitations which we have implemented long-term
#      workarounds for.

# fields with changing values, which make it hard to compare values directly
FICKLE_FIELDS = {
    'coupons': {
        'times_redeemed' # expect 0, get 1
    },
    'customers': set(),
    'subscriptions': set(),
    'products': set(),
    'invoice_items': set(),
    'payment_intents': set(),
    'payouts': {
        'object',      # expect 'transfer', get 'payout'
    },
    'charges': {
        'status',      # expect 'paid', get 'succeeded'
        'receipt_url', # keeps changing with every request
        'source',      # changes depending on source type
    },
    'subscription_items': set(),
    'invoices': {
        'hosted_invoice_url', # expect https://invoice.stripe.com/i/acct_14zvmQDcBSxinnbL/test...zcy0200wBekbjGw?s=ap
        'invoice_pdf',        # get    https://invoice.stripe.com/i/acct_14zvmQDcBSxinnbL/test...DE102006vZ98t5I?s=ap
        'payment_settings',   # 'default_mandate' subfield unexpectedly present
        'subscription_details'
    },
    'plans': set(),
    'invoice_line_items': set()
}

FIELDS_ADDED_BY_TAP = {
    'coupons': {'updated'},
    'customers': {'updated'},
    'subscriptions': {'updated'},
    'products': {'updated'},
    'invoice_items': {
        'updated',
        # BUG_13666 | [https://jira.talendforge.org/browse/TDL-13666]
        #             Deterimine what we do when creating records that
        #             cuases the presence of this value to be inconsistent
        'subscription_item',
    },
    'payouts': {'updated'},
    'charges': {'updated'},
    'subscription_items': set(), # `updated` is not added by the tap for child streams.
    'invoices': {'updated'},
    'plans': {'updated'},
    'payment_intents': {'updated'},
    'invoice_line_items': {
        'invoice'
    },
}

class ALlFieldsTest(BaseTapTest):
    """Test tap is able to successfully replicate all selected fields"""

    @staticmethod
    def name():
        return "tt_stripe_all_fields"

    def parse_bookmark_to_date(self, value):
        if value:
            if isinstance(value, str):
                return self.local_to_utc(parse(value))
            if isinstance(value, int):
                return self.local_to_utc(dt.utcfromtimestamp(value))
        return value

    @classmethod
    def setUpClass(cls):
        LOGGER.info("Start Setup")
        # Create data prior to first sync
        cls.streams_to_test = {
            "customers",
            "payment_intents",
            "charges",
            "coupons",
            "invoice_items",
            "invoice_line_items",
            "invoices",
            "payouts",
            "plans",
            "products",
            "subscription_items",
            "subscriptions",
        }

        cls.expected_objects = {stream: [] for stream in cls.streams_to_test}
        cls.existing_objects = {stream: [] for stream in cls.streams_to_test}
        cls.new_objects = {stream: [] for stream in cls.streams_to_test}

        for stream in cls.streams_to_test:

            # create new records
            stripe_obj = create_object(stream)
            cls.new_objects[stream] = [stripe_obj_to_dict(stripe_obj)]
            cls.expected_objects[stream] = cls.new_objects[stream]


    @classmethod
    def tearDownClass(cls):
        LOGGER.info("Start Teardown")
        for stream in cls.streams_to_test:
            for record in cls.new_objects[stream]:
                delete_object(stream, record["id"])


    def getPKsToRecordsDict(self, stream, records, duplicates=False):  # BUG_9720
        """Return dict object of tupled pk values to record"""
        primary_keys = list(self.expected_primary_keys().get(stream))

        if not duplicates: # just send back a dictionary comprehension of tupled pks to records
            pks_to_record_dict = {tuple(record.get(pk) for pk in primary_keys): record for record in records}
            return pks_to_record_dict, dict()

        # if duplicates are present we must track them in a separate dictionary
        pks_to_record_dict_1 = dict()
        pks_to_record_dict_2 = dict()
        for record in records:
            primary_key_values = tuple(record.get(pk) for pk in primary_keys)

            if pks_to_record_dict_1.get(primary_key_values):
                pks_to_record_dict_2[primary_key_values] = record
                continue

            pks_to_record_dict_1[primary_key_values] = record
        return pks_to_record_dict_1, pks_to_record_dict_2


    def test_run(self):

        # first just run the test against customers
        streams_to_test_1 = {'customers'}

        # then run against all streams under test (except customers)
        streams_to_test_2 = self.streams_to_test.difference(streams_to_test_1)

        for streams_to_test in [streams_to_test_1, streams_to_test_2]:
            with self.subTest(streams_to_test=streams_to_test):

                # get existing records and add them to our expectations
                for stream in streams_to_test:
                    stripe_obj = list_all_object(stream)
                    self.existing_objects[stream] = stripe_obj
                    self.expected_objects[stream] = self.existing_objects[stream]

                # run the test
                self.all_fields_test(streams_to_test)

    def all_fields_test(self, streams_to_test):
        """
        Verify that for each stream data is synced when all fields are selected.
        Verify the synced data matches our expectations based off of the applied schema
        and results from the test client utils.
        """

        # instantiate connection
        conn_id = connections.ensure_connection(self)
        self.conn_id = conn_id

        # run check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        self.perform_and_verify_table_and_field_selection(
            conn_id, found_catalogs, streams_to_test, select_all_fields=True
        )

        # run initial sync
        first_record_count_by_stream = self.run_and_verify_sync(conn_id)

        replicated_row_count = sum(first_record_count_by_stream.values())
        synced_records = runner.get_records_from_target_output()

        # Verify target has records for all synced streams
        for stream, count in first_record_count_by_stream.items():
            assert stream in self.expected_streams()
            self.assertGreater(count, 0, msg="failed to replicate any data for: {}".format(stream))
        LOGGER.info("total replicated row count: {}".format(replicated_row_count))


        # Test by Stream
        for stream in streams_to_test:
            with self.subTest(stream=stream):

                # set expectations
                primary_keys = list(self.expected_primary_keys().get(stream))
                expected_records = self.records_data_type_conversions(
                    self.expected_objects[stream]
                )
                expected_records_keys = set()
                for record in expected_records:
                    expected_records_keys.update(set(record.keys()))

                # collect actual values
                actual_records = synced_records.get(stream)
                # Get the actual stream records based on the newly added field `updated_by_event_type`
                # as the events endpoints is not the latest version and hence returns deprecated fields also.
                actual_record_message = actual_records.get('messages')
                actual_records_data = [message['data'] for message in actual_record_message
                                       if not message.get('data').get('updated_by_event_type', None)]

                actual_records_keys = set()
                for message in actual_record_message:
                    # Get the actual stream records which would have `updated_by_event_type` as None
                    if message['action'] == 'upsert' and not message.get('data').get('updated_by_event_type', None):
                        actual_records_keys.update(set(message['data'].keys()))
                schema_keys = set(self.expected_schema_keys(stream)) # read in from schema files

                # Get event based records based on the newly added field `updated_by_event_type`
                events_records_data = [message['data'] for message in actual_record_message
                                       if message.get('data').get('updated_by_event_type', None)]

                # To avoid warning, skipping fields of FIELDS_TO_NOT_CHECK
                schema_keys = schema_keys - FIELDS_TO_NOT_CHECK.get(stream, set())
                actual_records_keys = actual_records_keys - FIELDS_TO_NOT_CHECK[stream]
                expected_records_keys = expected_records_keys - FIELDS_TO_NOT_CHECK[stream]

                # Append fields which are added by tap to expectation
                adjusted_expected_keys = expected_records_keys.union(
                    FIELDS_ADDED_BY_TAP.get(stream, set())
                )

                # Log the fields that are included in the schema but not in the expectations.
                # These are fields we should strive to get data for in our test data set
                if schema_keys.difference(adjusted_expected_keys):
                    LOGGER.warn("Stream[{}] Fields missing from expectations: [{}]".format(
                        stream, schema_keys.difference(adjusted_expected_keys)
                    ))

                adjusted_actual_keys = actual_records_keys.union(  # BUG_12478
                    KNOWN_MISSING_FIELDS.get(stream, set())
                ).union(SCHEMA_MISSING_FIELDS.get(stream, set()))

                if stream == 'invoice_items':
                    adjusted_actual_keys = adjusted_actual_keys.union({'subscription_item'})  # BUG_13666

                # Verify the expected_keys is a subset of the actual_keys
                self.assertTrue(adjusted_expected_keys.issubset(adjusted_actual_keys), msg=f"{adjusted_expected_keys} is not a subset of {adjusted_actual_keys}")

                # verify the missing fields from KNOWN_MISSING_FIELDS are always missing (stability check)
                self.assertSetEqual(actual_records_keys.difference(KNOWN_MISSING_FIELDS.get(stream, set())), actual_records_keys)

                # Verify that all fields sent to the target fall into the expected schema
                self.assertTrue(actual_records_keys.issubset(schema_keys), msg=f"{actual_records_keys} is not a subset of {schema_keys}")

                # Verify there are no duplicate pks in the target
                actual_pks = [tuple(actual_record.get(pk) for pk in primary_keys) for actual_record in actual_records_data]
                actual_pks_set = set(actual_pks)

                # self.assertEqual(len(actual_pks_set), len(actual_pks))  # BUG_9720
                # assert unique primary keys for actual records
                self.assertLessEqual(len(actual_pks_set), len(actual_pks))

                # Verify there are no duplicate pks in our expectations
                expected_pks = [tuple(expected_record.get(pk) for pk in primary_keys) for expected_record in expected_records]
                expected_pks_set = set(expected_pks)
                self.assertEqual(len(expected_pks_set), len(expected_pks))

                # Get event-based pks based on the newly added field `updated_by_event_type` and verify
                # there are no duplicate pks in our expectations
                events_based_actual_pks = [tuple(event_record.get(pk) for pk in primary_keys) for event_record in events_records_data]
                events_based_actual_pks_set = set(events_based_actual_pks)

                # Verify unique primary keys for event-based records
                self.assertLessEqual(len(events_based_actual_pks_set), len(events_based_actual_pks))

                # Verify by pks, that we replicated the expected records
                self.assertTrue(actual_pks_set.issuperset(expected_pks_set))

                # test records by field values...
                expected_pks_to_record_dict, _ = self.getPKsToRecordsDict(stream, expected_records)  # BUG_9720
                actual_pks_to_record_dict, actual_pks_to_record_dict_dupes = self.getPKsToRecordsDict(  # BUG_9720
                    stream, actual_records_data, duplicates=True
                )

                # BUG_9720 | https://jira.talendforge.org/browse/TDL-9720

                # Verify the fields which are replicated, adhere to the expected schemas
                for pks_tuple, expected_record in expected_pks_to_record_dict.items():
                    with self.subTest(record=pks_tuple):

                        actual_record = actual_pks_to_record_dict.get(pks_tuple) or {}

                        # BUG_9720 | uncomment to reproduce a duplicate record with a data discrepancy
                        # actual_record_dupe = actual_pks_to_record_dict_dupes.get(pks_tuple) or {}
                        # if actual_record_dupe != actual_record and \
                        #    actual_record_dupe['created'] == actual_record['created'] and \
                        #    actual_record_dupe['updated'] == actual_record['updated']:
                        #     import pdb; pdb.set_trace()
                        #     LOGGER.info(f"Discrepancy {set(actual_record_dupe.keys()).difference(set(actual_record.keys())))}")
                        #     LOGGER.info("created: {actual_record['created']}")
                        #     LOGGER.info("created dupe: {actual_record_dupe['created']}")
                        #     LOGGER.info("updated: {actual_record['updated']}")
                        #     LOGGER.info("updated dupe: {actual_record_dupe['updated']}")

                        field_adjustment_set = FIELDS_ADDED_BY_TAP[stream].union(
                            KNOWN_MISSING_FIELDS.get(stream, set())  # BUG_12478
                        )

                        # NB | THere are many subtleties in the stripe Data Model.

                        #      We have seen multiple cases where Field A in Stream A has an effect on Field B in Stream B.
                        #      Stripe also appears to run frequent background processes which can result in the update of a
                        #      record between the time when we set our expectations and when we run a sync, therefore
                        #      invalidating our expectations.

                        #      To work around these challenges we will attempt to compare fields directly. If that fails
                        #      we will log the inequality and assert that the datatypes at least match.

                        for field in set(actual_record.keys()).difference(field_adjustment_set):  # skip known bugs
                            with self.subTest(field=field):
                                base_err_msg = f"Stream[{stream}] Record[{pks_tuple}] Field[{field}]"

                                expected_field_value = expected_record.get(field, "EXPECTED IS MISSING FIELD")
                                actual_field_value = actual_record.get(field, "ACTUAL IS MISSING FIELD")

                                try:

                                    self.assertEqual(expected_field_value, actual_field_value)

                                except AssertionError as failure_1:

                                    LOGGER.warn(f"{base_err_msg} failed exact comparison.\n"
                                        f"AssertionError({failure_1})")

                                    if field in KNOWN_FAILING_FIELDS[stream] or field in FIELDS_TO_NOT_CHECK[stream]:
                                        continue # skip the following wokaround

                                    elif actual_field_value and field in FICKLE_FIELDS[stream]:
                                        self.assertIsInstance(actual_field_value, type(expected_field_value))

                                    elif actual_field_value:
                                        raise AssertionError(f"{base_err_msg} Unexpected field is being fickle.")

                                    else:
                                        LOGGER.warn(f"{base_err_msg} failed datatype comparison. Field is None.")



from utils import stripe_obj_to_dict, client, midnight
from datetime import datetime as dt
from datetime import timedelta, time
from tap_tester import runner, connections
from base import BaseTapTest

def get_payouts():
    """
        Return all the payouts (with pagination), to determine the automatic and non-automatic payouts
    """
    # list of all data to return
    four_days_ago = int(dt.combine(dt.today()-timedelta(days=4), time.min).timestamp())
    data = []
    # Api call of 1st page starting from 4 days ago as there is a lag from the Stripe side to reflect
    # the automatic payout transactions data
    stripe_obj = client["payouts"].list(limit=100, created={"gte": four_days_ago})
    dict_obj = stripe_obj_to_dict(stripe_obj)

    try:
        # add data
        data += dict_obj['data']
    except KeyError:
        raise Exception("No records for 'Payouts' were replicated, please run 'test_all_fields' before re-running.")

    # loop over rest of the pages and collect data
    while dict_obj.get("has_more"):
        stripe_obj = client["payouts"].list(limit=100, created={"gte": four_days_ago}, starting_after=dict_obj.get('data')[-1].get('id'))
        dict_obj = stripe_obj_to_dict(stripe_obj)
        data += dict_obj['data']

    # send data
    return data

class AutomaticPayoutTransactionTest(BaseTapTest):
    """
        Test case to verify that we only collect payout_transactions for payouts containing "automatic" field as "True"
        Prerequisite:
            Run 'test_all_fields' before running this test case.
    """

    @staticmethod
    def name():
        return "tt_stripe_automatic_payout_transactions"

    @classmethod
    def setUpClass(cls):
        """
            Create list of automatic and non-automatic payouts used for assertion
        """
        # get all the payouts
        payouts = get_payouts()

        # create lists of payout ids containing automatic field as "True" and "False"
        cls.payouts_with_automatic_true = []
        cls.payouts_with_automatic_false = []
        for record in payouts:
            if record.get("automatic"):
                cls.payouts_with_automatic_true.append(record.get("id"))
            else:
                cls.payouts_with_automatic_false.append(record.get("id"))

    def test_run(self):
        # Decreased the start_date for payout_transactions stream as there is a lag from the Stripe side to reflect
        # the automatic payout transactions data
        self.start_date = dt.strftime(dt.today() - timedelta(days=4), self.START_DATE_FORMAT)
        conn_id = connections.ensure_connection(self, original_properties=False)

        expected_streams = {"payouts", "payout_transactions"}

        # Select payouts and payout_transactions streams
        found_catalogs = self.run_and_verify_check_mode(conn_id)
        our_catalogs = [catalog for catalog in found_catalogs if
                        catalog.get("tap_stream_id") in expected_streams]

        # field selection
        self.select_all_streams_and_fields(conn_id, our_catalogs)

        # Run a sync job using orchestrator
        first_sync_record_count = self.run_and_verify_sync(conn_id)

        # Get the set of records from a first sync
        first_sync_records = runner.get_records_from_target_output()

        # set stream as "payout_transactions"
        stream = "payout_transactions"
        with self.subTest(stream=stream):
            # verify that there is more than 1 record
            self.assertGreater(first_sync_record_count.get(stream, -1), 0,
                               msg="Data isn't set up to be able to test full sync")

            # get records
            records = [message.get("data") for message in first_sync_records.get(stream).get("messages") if message["action"] == "upsert"]

            # collect payout ids for all the payout transaction records
            payout_transaction_payout_ids = set()
            for record in records:
                payout_transaction_payout_ids.add(record.get("payout_id"))

            # verify that data exists for payouts with "automatic" field as "True" and "False"
            self.assertTrue(self.payouts_with_automatic_true is not None)
            self.assertTrue(self.payouts_with_automatic_false is not None)

            # loop over all the payout ids from the payout transactions to verify
            # that we collected "payout transactions" of automatic payouts only
            for id in payout_transaction_payout_ids:
                # verify that we collect payout transaction record for payout containing "automatic": True
                self.assertTrue(id in self.payouts_with_automatic_true)
                # verify that we do not collect payout transaction record for payout containing "automatic": False
                self.assertTrue(id not in self.payouts_with_automatic_false)



"""
Test that the start_date configuration is respected
"""

from functools import reduce
from datetime import datetime as dt
from datetime import timedelta
from dateutil.parser import parse

from tap_tester import menagerie, runner, connections

from base import BaseTapTest
from utils import create_object, update_object, update_payment_intent, delete_object, get_catalogs


class StartDateTest(BaseTapTest):
    """
    Test that the start_date configuration is respected

    • verify that a sync with a later start date has at least one record synced
      and less records than the 1st sync with a previous start date
    • verify that each stream has less records than the earlier start date sync
    • verify all data from later start data has bookmark values >= start_date
    • verify that the minimum bookmark sent to the target for the later start_date sync
      is greater than or equal to the start date
    • verify by primary key values, that all records in the 1st sync are included in the 2nd sync.
    """

    @staticmethod
    def name():
        return "tt_stripe_start_date"

    def test_run(self):
        """Test we get a lot of data back based on the start date configured in base"""
        conn_id = connections.ensure_connection(self)
        self.conn_id = conn_id

        # Select all streams and all fields within streams
        found_catalogs = self.run_and_verify_check_mode(conn_id)
        incremental_streams = {key for key, value in self.expected_replication_method().items()
                               if value == self.INCREMENTAL}

        # IF THERE ARE STREAMS THAT SHOULD NOT BE TESTED
        # REPLACE THE EMPTY SET BELOW WITH THOSE STREAMS
        untested_streams = self.child_streams().union({
            'events',
            'payout_transactions'
        })
        our_catalogs = get_catalogs(conn_id, incremental_streams.difference(untested_streams))

        self.select_all_streams_and_fields(conn_id, our_catalogs, select_all_fields=True)

        # Create a record for each stream under test prior to the first sync
        new_objects = {
            stream: create_object(stream)
            for stream in incremental_streams.difference(untested_streams)
        }

        # Run a sync job using orchestrator
        first_sync_record_count = self.run_and_verify_sync(conn_id)
        first_total_records = reduce(lambda a, b: a + b, first_sync_record_count.values())

        # Count actual rows synced
        first_sync_records = runner.get_records_from_target_output()
        first_sync_created, _ = self.split_records_into_created_and_updated(first_sync_records)

        # set the start date for a new connection based off bookmarks largest value
        first_max_bookmarks = self.max_bookmarks_by_stream(first_sync_records)

        bookmark_list = [next(iter(book.values())) for stream, book in first_max_bookmarks.items()]
        bookmark_dates = []
        for bookmark in bookmark_list:
            try:
                bookmark_dates.append(parse(bookmark))
            except (ValueError, OverflowError, TypeError):
                pass

        if not bookmark_dates:
            # THERE WERE NO BOOKMARKS THAT ARE DATES.
            # REMOVE CODE TO FIND A START DATE AND ENTER ONE MANUALLY
            raise ValueError

        # largest_bookmark = reduce(lambda a, b: a if a > b else b, bookmark_dates)
        # self.start_date = self.local_to_utc(largest_bookmark).strftime(self.START_DATE_FORMAT)

        self.start_date = dt.strftime(dt.today() - timedelta(days=1), self.START_DATE_FORMAT)

        # create a new connection with the new start_date

        conn_id = connections.ensure_connection(self, original_properties=False)
        self.conn_id = conn_id

        # Select all streams and all fields within streams
        found_catalogs = self.run_and_verify_check_mode(conn_id)
        our_catalogs = [catalog for catalog in found_catalogs if
                        catalog.get('tap_stream_id') in incremental_streams.difference(
                            untested_streams)]
        self.select_all_streams_and_fields(conn_id, our_catalogs, select_all_fields=True)

        # Update a record for each stream under test prior to the 2nd sync
        updated = {}  # holds id for updated objects in each stream
        for stream in new_objects:
            if stream == 'payment_intents':
                # updating the PaymentIntent object may require multiple attempts
                record = update_payment_intent(stream)
            else:
                # There needs to be some test data for each stream, otherwise this will break
                record = first_sync_created[stream]["messages"][0]["data"]
                update_object(stream, record["id"])
            updated[stream] = record["id"]

        # Run a sync job using orchestrator
        second_sync_record_count = self.run_and_verify_sync(conn_id, clear_state=True)

        # tap-stripe uses events for updates, so these need filtered to validate bookmark
        second_sync_records = runner.get_records_from_target_output()
        second_sync_created, second_sync_updated = self.split_records_into_created_and_updated(second_sync_records)
        second_total_records = reduce(lambda a, b: a + b, second_sync_record_count.values(), 0)

        # Only examine bookmarks for "created" objects, not updates
        second_min_bookmarks = self.min_bookmarks_by_stream(second_sync_created)

        # verify that at least one record synced and less records synced than the 1st connection
        self.assertGreater(second_total_records, 0)
        self.assertLess(first_total_records, second_total_records)
        stream_primary_keys = self.expected_primary_keys()

        # validate that all newly created records are greater than the start_date
        for stream in incremental_streams.difference(untested_streams):
            with self.subTest(stream=stream):

                # verify that each stream has more or equal records in the second sync than in the first
                self.assertGreaterEqual(
                    second_sync_record_count.get(stream, 0),
                    first_sync_record_count.get(stream, 0),
                    msg="first had more records, start_date usage not verified")

                # verify all data from 2nd sync >= start_date
                target_mark = second_min_bookmarks.get(stream, {"mark": None})
                target_value = next(iter(target_mark.values()))  # there should be only one

                record_count_1 = first_sync_record_count[stream]
                record_count_2 = second_sync_record_count[stream]
                primary_keys_list_1 = [tuple(message.get('data').get(expected_pk) for expected_pk in stream_primary_keys[stream])
                                       for message in first_sync_records.get(stream).get('messages')
                                       if message.get('action') == 'upsert']
                primary_keys_list_2 = [tuple(message.get('data').get(expected_pk) for expected_pk in stream_primary_keys[stream])
                                       for message in second_sync_records.get(stream).get('messages')
                                       if message.get('action') == 'upsert']
                primary_keys_sync_1 = set(primary_keys_list_1)
                primary_keys_sync_2 = set(primary_keys_list_2)

                # Verify by primary key values, that all records in the 1st sync are included in the 2nd sync.
                self.assertTrue(
                        primary_keys_sync_1.issubset(primary_keys_sync_2))

                if target_value:

                    # it's okay if there isn't target data for a stream
                    try:
                        target_value = self.local_to_utc(parse(target_value))
                        expected_value = self.local_to_utc(parse(self.start_date))
                        # verify that the minimum bookmark sent to the target for the second sync
                        # is greater than or equal to the start date

                        # TODO - BUG https://jira.talendforge.org/browse/TDL-20911
                        # There is a lookback window being applied to the start_date, but the lookback window should not go beyond the startdate
                        if stream in ('balance_transactions', 'events'):
                            expected_value = expected_value - timedelta(minutes=10)
                        self.assertGreaterEqual(target_value, expected_value)

                    except (OverflowError, ValueError, TypeError):
                        LOGGER.warn("bookmarks cannot be converted to dates, can't test start_date for %s", stream)

                if stream in updated:
                    delete_object(stream, updated[stream])



"""
Test tap gets all creates for streams (as long as we can create an object)
"""

from tap_tester import menagerie, runner, connections, LOGGER
from base import BaseTapTest
from utils import create_object, delete_object


class CreateObjectTest(BaseTapTest):
    """Test tap gets all creates for streams (as long as we can create an object)"""

    @staticmethod
    def name():
        return "tt_stripe_create_objects"

    def test_run(self):
        """
        Verify that the sync only sent records to the target for selected streams
        Create a new object for each stream
        Verify that the second sync includes at least one create for each stream
        Verify that the created record was picked up on the second sync
        """
        conn_id = connections.ensure_connection(self)
        self.conn_id = conn_id

        streams_to_create = {
            "balance_transactions",  # should be created implicity with a create in the payouts or charges streams
            "charges",
            "coupons",
            "customers",
            "invoice_items",
            "invoice_line_items", # this is created implicity by invoices, it just creates another invoice TODO get this outa here
            "invoices", # this will create an invoice_item
            "payouts",
            "plans",
            "payment_intents",
            "products",
            "subscription_items",
            "subscriptions", # this will create a new plan and payment method
         }

        missing_streams_to_create = {
            "disputes",  # can be created by simulating a dispute transaction with a specific card number
            # no way to create directly, see: https://stripe.com/docs/testing#disputes
            "payout_transactions",  # BUG_9703 | https://jira.talendforge.org/browse/TDL-9703
            # depends on payouts and transactions
            "transfers",
            # needs an account that we can transfer to, not sure
            # how to set up a test account we can use to create a transfer
         }

        our_catalogs = self.run_and_verify_check_mode(conn_id)

        self.select_all_streams_and_fields(
            conn_id, our_catalogs, select_all_fields=True
        )

        # Run a sync job using orchestrator
        first_sync_record_count = self.run_and_verify_sync(conn_id)

        # verify that the sync sent records to the target for selected streams (catalogs)
        self.assertTrue(streams_to_create.issubset(set(first_sync_record_count.keys())))

        # Get the set of records from a first sync
        first_sync_records = runner.get_records_from_target_output()

        first_sync_created, _ = self.split_records_into_created_and_updated(
            first_sync_records
        )

        new_objects = {
            stream: create_object(stream)
            for stream in streams_to_create.difference({"balance_transactions"})
        }

        # Run a second sync job using orchestrator
        second_sync_record_count = self.run_and_verify_sync(conn_id)

        # Get the set of records from a second sync
        second_sync_records = runner.get_records_from_target_output()

        second_sync_created, _ = self.split_records_into_created_and_updated(
            second_sync_records
        )

        # # THIS MAKES AN ASSUMPTION THAT CHILD STREAMS DO NOT NEED TESTING.
        # # ADJUST IF NECESSARY
        for stream in streams_to_create.difference(self.child_streams()):
            with self.subTest(stream=stream):

                second_sync_created_objects = second_sync_created.get(stream, {}).get(
                    "messages", []
                )

                # verify that you get at least one new record on the second sync
                self.assertGreaterEqual(
                    len(second_sync_created_objects),
                    1,
                    msg="second sync didn't have created objects",
                )

                if stream == "balance_transactions":
                    sources = [record.get("data", {}).get("source")
                               for record in second_sync_created_objects]

                    self.assertTrue(new_objects['payouts']['id'] in sources)
                    self.assertTrue(new_objects['charges']['id'] in sources)

                    continue

                # TODO START DEBUG
                # remove debug after BUG https://jira.talendforge.org/browse/TDL-21614 is resolved
                if stream == 'invoices':
                    null_date_invoices = []
                    masking_invoices = []
                    for rec in second_sync_records[stream]['messages']:
                        # detect old failures by comparing record dates using both replication keys
                        # it is believed that the created invoices never have 'date' and should
                        # always fail verification due to the old split logic
                        if not rec['data'].get('date'):
                            if rec['data'].get('created') == rec['data'].get('updated'):
                                null_date_invoices += [rec['data']['id']]
                        # date key was found for records in the else clause.  It is believed that
                        # these are all updated records. Check to see if failure would be masked
                        # by the split logic
                        else:
                            if rec['data'].get('date') == rec['data'].get('updated'):
                                masking_invoices += [rec['data']['id']]
                    LOGGER.info(f"null_date_invoices: {null_date_invoices}, "
                                f"masking_invoices: {masking_invoices}, "
                                f"new_id: {new_objects[stream]['id']}")
                    self.assertTrue(new_objects[stream]['id'] in null_date_invoices)
                    if new_objects[stream]['id'] not in masking_invoices:
                        LOGGER.warn(f"########## Previous error scenario detected (un-masked failure) ##########")
                # TODO END DEBUG

                # verify the new object is in the list of created objects
                # from the second sync
                self.assertTrue(
                    any(
                        new_objects[stream]["id"] == record.get("data", {}).get("id")
                        for record in second_sync_created_objects
                    )
                )

                if stream in streams_to_create:
                    delete_object(stream, new_objects[stream]["id"])



"""
Test that with no fields selected for a stream automatic fields are still replicated
"""

from tap_tester import runner, menagerie, connections

from base import BaseTapTest
from utils import create_object


class MinimumSelectionTest(BaseTapTest):
    """Test that with no fields selected for a stream automatic fields are still replicated"""

    @staticmethod
    def name():
        return "tt_stripe_auto_fields"

    def test_run(self):
        """
        Verify that for each stream you can get multiple pages of data
        when no fields are selected and only the automatic fields are replicated.

        PREREQUISITE
        For EACH stream add enough data that you surpass the limit of a single
        fetch of data.  For instance if you have a limit of 250 records ensure
        that 251 (or more) records have been posted for that stream.
        """
        conn_id = connections.ensure_connection(self)
        self.conn_id = conn_id
        streams_to_create = {
            # "balance_transactions",  # should be created implicity with a create in the payouts or charges streams
            "charges",
            "coupons",
            "customers",
            "invoice_items",
            "invoice_line_items", # this is created implicity by invoices, it just creates another invoice
            "invoices", # this will create an invoice_item
            "payouts",
            "plans",
            "payment_intents",
            "products",
            "subscription_items",
            "subscriptions", # this will create a new plan and payment method
            "transfers",
         }
        untested_streams = {
            "payout_transactions",
            "disputes"
        }
        new_objects = {
            stream: create_object(stream)
            for stream in streams_to_create.difference()
        }


        # Select all streams and no fields within streams
        # IF THERE ARE NO AUTOMATIC FIELDS FOR A STREAM
        # WE WILL NEED TO UPDATE THE BELOW TO SELECT ONE
        found_catalogs = self.run_and_verify_check_mode(conn_id)
        self.select_all_streams_and_fields(conn_id, found_catalogs, select_all_fields=False)

        # Run a sync job using orchestrator
        record_count_by_stream = self.run_and_verify_sync(conn_id)

        actual_fields_by_stream = runner.examine_target_output_for_fields()

        for stream in self.expected_streams().difference(untested_streams):
            with self.subTest(stream=stream):

                # verify that you get some records for each stream
                # SKIP THIS ASSERTION FOR STREAMS WHERE YOU CANNOT GET
                # MORE THAN 1 PAGE OF DATA IN THE TEST ACCOUNT
                self.assertGreater(
                    record_count_by_stream.get(stream, -1), 0,
                    msg="The number of records is not over the stream max limit")

                # verify that only the automatic fields are sent to the target
                actual = actual_fields_by_stream.get(stream) or set()
                expected = self.expected_automatic_fields().get(stream, set())
                self.assertEqual(
                    actual, expected,
                    msg=("The fields sent to the target are not the automatic fields. Expected: {}, Actual: {}"
                         .format(actual, expected))
                )



"""
Test tap discovery
"""
import re

from tap_tester import menagerie, connections

from base import BaseTapTest


class DiscoveryTest(BaseTapTest):
    """ Test the tap discovery """

    @staticmethod
    def name():
        return "tt_stripe_discovery"

    def test_run(self):
        """
        Verify that discover creates the appropriate catalog, schema, metadata, etc.

        • Verify number of actual streams discovered match expected
        • Verify the stream names discovered were what we expect
        • Verify stream names follow naming convention
          streams should only have lowercase alphas and underscores
        • verify there is only 1 top level breadcrumb
        • verify replication key(s)
        • verify primary key(s)
        • verify that if there is a replication key we are doing INCREMENTAL otherwise FULL
        • verify the actual replication matches our expected replication method
        • verify that primary, replication and foreign keys
          are given the inclusion of automatic (metadata and annotated schema).
        • verify that all other fields have inclusion of available (metadata and schema)
        • verify there are no duplicate metadata entries
        """
        conn_id = connections.ensure_connection(self)
        self.conn_id = conn_id

        # Verify number of actual streams discovered match expected
        found_catalogs = self.run_and_verify_check_mode(conn_id)
        self.assertGreater(len(found_catalogs), 0,
                           msg="unable to locate schemas for connection {}".format(conn_id))
        self.assertEqual(len(found_catalogs),
                         len(self.expected_streams()),
                         msg="Expected {} streams, actual was {} for connection {},"
                             " actual {}".format(
                                 len(self.expected_streams()),
                                 len(found_catalogs),
                                 found_catalogs,
                                 conn_id))

        # Verify the stream names discovered were what we expect
        found_catalog_names = {c['tap_stream_id'] for c in found_catalogs}
        self.assertEqual(set(self.expected_streams()),
                         set(found_catalog_names),
                         msg="Expected streams don't match actual streams")

        # Verify stream names follow naming convention
        # streams should only have lowercase alphas and underscores
        self.assertTrue(all([re.fullmatch(r"[a-z_]+", name) for name in found_catalog_names]),
                        msg="One or more streams don't follow standard naming")

        for stream in self.expected_streams():
            with self.subTest(stream=stream):
                catalog = next(iter([catalog for catalog in found_catalogs
                                     if catalog["stream_name"] == stream]))
                assert catalog  # based on previous tests this should always be found

                schema_and_metadata = menagerie.get_annotated_schema(conn_id, catalog['stream_id'])
                metadata = schema_and_metadata["metadata"]
                schema = schema_and_metadata["annotated-schema"]

                # verify the stream level properties are as expected
                # verify there is only 1 top level breadcrumb
                stream_properties = [item for item in metadata if item.get("breadcrumb") == []]
                self.assertTrue(len(stream_properties) == 1,
                                msg="There is more than one top level breadcrumb")

                actual_fields = [md_entry.get("breadcrumb")[1] for md_entry in metadata if md_entry.get("breadcrumb") != []]

                # verify there are no duplicate metadata entries
                self.assertEqual(len(actual_fields), len(set(actual_fields)), msg = f"duplicates in the fields retrieved")
                
                # verify replication key(s)
                actual = set(stream_properties[0].get("metadata", {self.REPLICATION_KEYS: []}).get(self.REPLICATION_KEYS) or [])
                expected = self.expected_replication_keys()[stream] or set()
                self.assertEqual(
                    actual,
                    expected,
                    msg="expected replication key {} but actual is {}".format(
                        expected, actual))

                # verify primary key(s)
                self.assertEqual(
                    set(stream_properties[0].get(
                        "metadata", {self.PRIMARY_KEYS: []}).get(self.PRIMARY_KEYS, [])),
                    self.expected_primary_keys()[stream],
                    msg="expected primary key {} but actual is {}".format(
                        self.expected_primary_keys()[stream],
                        set(stream_properties[0].get(
                            "metadata", {self.PRIMARY_KEYS: None}).get(self.PRIMARY_KEYS, []))))

                # verify that if there is a replication key we are doing INCREMENTAL otherwise FULL
                actual_replication_method = stream_properties[0].get(
                    "metadata", {self.REPLICATION_METHOD: None}).get(self.REPLICATION_METHOD)

                # invoice_line_items has no bookmark, but is an incremental child of invoices
                if stream == 'invoice_line_items' or stream_properties[0].get(
                        "metadata", {self.REPLICATION_KEYS: []}).get(self.REPLICATION_KEYS, []):

                    self.assertTrue(actual_replication_method == self.INCREMENTAL,
                                    msg="Expected INCREMENTAL replication "
                                        "since there is a replication key")
                else:
                    self.assertTrue(actual_replication_method == self.FULL,
                                    msg="Expected FULL replication "
                                        "since there is no replication key")

                # verify the actual replication matches our expected replication method
                self.assertEqual(
                    self.expected_replication_method().get(stream, None),
                    actual_replication_method,
                    msg="The actual replication method {} doesn't match the expected {}".format(
                        actual_replication_method,
                        self.expected_replication_method().get(stream, None)))

                expected_automatic_fields = self.expected_automatic_fields()[stream] or set()

                # verify that primary, replication and foreign keys
                # are given the inclusion of automatic in annotated schema.
                actual_automatic_fields = {key for key, value in schema["properties"].items()
                                           if value.get("inclusion") == "automatic"}
                self.assertEqual(expected_automatic_fields,
                                 actual_automatic_fields,
                                 msg="expected {} automatic fields but got {}".format(
                                     expected_automatic_fields,
                                     actual_automatic_fields))

                # verify that all other fields have inclusion of available
                # This assumes there are no unsupported fields for SaaS sources
                self.assertTrue(
                    all({value.get("inclusion") == "available" for key, value
                         in schema["properties"].items()
                         if key not in actual_automatic_fields}),
                    msg="Not all non key properties are set to available in annotated schema")

                # verify that primary, replication and foreign keys
                # are given the inclusion of automatic in metadata.
                actual_automatic_fields = \
                    {item.get("breadcrumb", ["properties", None])[1]
                     for item in metadata
                     if item.get("metadata").get("inclusion") == "automatic"}
                self.assertEqual(expected_automatic_fields,
                                 actual_automatic_fields,
                                 msg="expected {} automatic fields but got {}".format(
                                     expected_automatic_fields,
                                     actual_automatic_fields))

                # verify that all other fields have inclusion of available
                # This assumes there are no unsupported fields for SaaS sources
                self.assertTrue(
                    all({item.get("metadata").get("inclusion") == "available"
                         for item in metadata
                         if item.get("breadcrumb", []) != []
                         and item.get("breadcrumb", ["properties", None])[1]
                         not in actual_automatic_fields}),
                    msg="Not all non key properties are set to available in metadata")





from tap_tester import runner, connections
from datetime import datetime as dt
from datetime import timedelta
from base import BaseTapTest

class ParentChildIndependentTest(BaseTapTest):

    @staticmethod
    def name():
        return "tt_stripe_parent_child_test"

    def test_child_streams(self):
        """
            Test case to verify that tap is working fine if only first level child streams are selected
        """
        four_days_ago = dt.strftime(dt.today() - timedelta(days=4), self.START_DATE_FORMAT)
        # select child streams only and run the test
        child_streams = {"invoice_line_items", "subscription_items"}
        self.run_test(child_streams)
        # Separated the payout_transactions stream as there is a lag from the Stripe side to reflect
        # the automatic payout transactions data, hence we want to change the start_date for that stream
        child_streams = {"payout_transactions"}
        start_date = four_days_ago
        self.run_test(child_streams, start_date, False)

    def run_test(self, streams, start_date=None, default_start_date=True):
        """
            Testing that tap is working fine if only child streams are selected
            - Verify that if only child streams are selected then only child stream are replicated.
        """

        if not default_start_date:
            self.start_date = start_date
        # instantiate connection
        conn_id = connections.ensure_connection(self, original_properties=default_start_date)
        self.conn_id = conn_id

        # run check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        self.perform_and_verify_table_and_field_selection(conn_id, found_catalogs, streams_to_select=streams)

        # run initial sync
        record_count_by_stream = self.run_and_verify_sync(conn_id)
        synced_records = runner.get_records_from_target_output()

        # Verify no unexpected streams were replicated
        synced_stream_names = set(synced_records.keys())
        self.assertSetEqual(streams, synced_stream_names)
